// ================================================================
// STEM Starter project
// using: STEM / Educational Game Platform - using a single top level Menu selection
// with a Micro Controller, 8 LED 4 color bar and 4 Button module + piezo sounder
//
// license: Attribution-NonCommercial-ShareAlike
// Aug-Sep 2020 Ron Miller, Initial Development,  late 2023 Further Development

//  include Library of Hardware Support functions
#include "4Btn_8LED_Platform.h"

#include <EEPROM.h>
// the address in the EEPROM we're going to write to
int eeCntAddr = 77;  // totally arbitary

//#define DEBUG false
bool DEBUG = false;
bool spkr = true; // speakers & pasive piezos can be AC driven for more volume (not active buzzers)
bool quiet = false; // assumes a 150 ohm resister between AUDIO_NEG & GND

// ------------------------ function prototypes
void showLevel(byte);
byte get_Mode(byte);
byte get_Option(byte, int);
bool YesNo();
bool btnDown();
void stableBtns();
void SimonSays();

// global game stuff
bool playing;
int turn;  // turn counter
byte ver;
byte level, goal;
unsigned int idleCnt=0, loopCnt;

unsigned long t0,t1;
unsigned long randSeed = 0;

bool volHigh = false;

//   LED display stuff
int cursorPtr;
bool colorLeds[5];
const int colorSet[4][2] = { {1, 2},  {3, 4},  {5, 6},  {7, 8} };
byte cell[14];  // working buffer


// the setup routine runs once on power up or when you press reset:
//=================================================================
void setup() {

  // ------------ setup the Mini Game/STEM platform (8LEDs, 4Btns, and Audio)
  gPlatform_setup();

  // initialize serial communication
  Serial.begin(115200);
  Serial.println("----------------------- STEM Starter sketch ");

  // ------ setup optional behavior based on button states at startup
  stableBtns(); // get btns state after debouncing
  if (Btn) {
    // WAS: ahead btn1 held at startup disables Farnsworth
    quiet = btn1; //=false; // now: enable quiet opperation
    //if (btn2)
    if (btn3) randomSeed(randSeed=314159); // always use the same rand #s for each run
    DEBUG = btn4;
    // note btn4 toggles Loud during Main menu
    Serial.print("button state at startup: "); 
    Serial.print(btn1); Serial.print(btn2); Serial.print(btn3); Serial.println(btn4);
  }
  if (quiet) pinMode(AUDIO_NEG, INPUT);
  
  delay(200);
  //  demonstrate there is Life in the components (LEDs)
  for (int i=0;i<4;i++) { refreshWait(100); set_LEDs(1<<i,1); set_LEDs(0x80>>i,1);}
  for (int i=4;i<8;i++) { refreshWait(100); set_LEDs(1<<i,2); set_LEDs(0x80>>i,2);}
  for (int i=0;i<4;i++) { refreshWait(100); set_LEDs(1<<i,3); set_LEDs(0x80>>i,3);}
  for (int i=4;i<8;i++) { refreshWait(100); set_LEDs(1<<i,0); set_LEDs(0x80>>i,0);}
  refreshWait(500);
  clearDisp();
 
  // highlight the Audio
  boop(40,250);
  beep(40,250);
  buZZ(40,250);
  //
  //LOUDbeep(30); delay(200); LOUDbuZZ(30);
  
//  delay(1000);
}

byte func;
// the loop routine runs over and over again forever:
//=================================================================
void loop() { // Top Menu level, for selecting 1 of N activities
  byte maxFunc;

  while(Btn) scanBtns();  // clear btns
  func=1;  // initial selection
  while(1) {
    clearDisp();
    while(Btn) {buZZ(1); scanBtns();}  // wait for no button

    maxFunc=7;  // currently there are 7 activities/games supported
    // indicate selectable functions
    for (int i=1; i<=maxFunc; i++) {dim[i]=true;}
    brightOne=func;
    refreshDisp();
    
    // watch for btn presses other than btn1
    while (! btn1) {
      refreshDisp();
      delay(1);
      scanBtns();
      //if (btn2Pressed) brightOne = (brightOne==1)?  maxFunc : brightOne-1;  // -- move left
      if (btn2Pressed) brightOne = (brightOne==0)?  maxFunc : brightOne-1;  // -- move left Allow 0:debugging
      if (btn3Pressed) brightOne = (brightOne%maxFunc) + 1; // ++ moves right
      // --------------- cycle: quiet-normal-Loud volume
      if (btn4Pressed) {
        if (Loud) { Loud=false; quiet=true;}
        else if (quiet) quiet=false;
        else Loud=true;
      }
      if (Loud && !spkr) {Loud=false; quiet=true;}
      if (quiet) pinMode(AUDIO_NEG, INPUT);
      else pinMode(AUDIO_NEG, OUTPUT);
      while (btn4) { genTone(650, 20); scanBtns();}
      //----------------
    }

    // randomize driven by real-world user event
    if (randSeed==0) {
      randomSeed(randSeed = micros()/4);  // micros() resolution is ~ 4us
      // used for QA testing: randomSeed(randSeed=36256276);  // starts with Rx test Ln=8 (7 w/ opt<3)
     Serial.print("random Seed:  "); Serial.print( randSeed); 
    }

    func = brightOne;
    clearDisp();
    while(Btn) {refreshWait(10); scanBtns();} // wait for button to clear

    Serial.print("\n FUNC: "); Serial.print(func);
    switch (func) {
      case 0:
        btnColor();
        while(btnDown());
        OneTimeTest();
        Serial.println();
        //btnReporting();    // exercises the buttons and Leds, and reports via Serial
        func=1;
        break;
      case 1:
        //PushIt();
        Blinker();     // Simple car turn signal blinker app, using low level support utilities.
        break;
      case 2:
        HW_Blinker(); // Simple car turn signal blinker app, using direct hardware I/O commands.
        // Or replace this with your own previously devised activity
        break;
      case 3:
        SuperComp60(); // simulate a 1960 super computer display board via blinking LEDs
        break;
      case 4:
        Skee_ball();  // A game of rolling “Balls” into desired spots.
        break;
      case 5:
        React_Now_ver2();      // Test your Reaction time
        break;
      case 6:
        SimpleSimon();   // A code simplified version of SimonSays
        break;
      case 7:
        pingPong();     // Two person game of Ping-Pong
        break;
      case 8:
        break;
      default:
        buZZ(1000);
    }
  }
}


void OneTimeTest() {
  // setup varied display
  clearDisp();
  dim[6]=dim[8]=1;
  lit[2]=lit[4]=1;
  brt[1]=brt[3]=1;
  flashLed=5;
  dimFlash=7;
  
  while(!btn4) {
    refreshWait(1);
    scanBtns();
    if (btn1 && BtnChanged) {LOUDbeep(150); beep(150); boop(150);}
    if (btn2 && BtnChanged) buZZ(200);
    if (btn3 && BtnChanged) LOUDbuZZ(200);
  }
}

void Blinker() {  // utilizes the the low level support layer found in Mini_STEM_Platform.h
    clearDisp();

    while (!(btn1 && btn4)) {
      refreshWait(1);
      scanBtns();
      if (btn1Pressed) {
        // note only 1 LED can have the 'flash' status
        flashLed = (flashLed==1)? 0 : 1;  // toggle the flashing
      }
      if (btn4Pressed) {
        flashLed = (flashLed==nleds)? 0 : nleds;
      }
    } 
}

void HW_Blinker() {  // implements 'Blinker' with direct low level Hardware control
  int i, Lite;
  byte btn1=0, btn4=0;
  byte lastB1, lastB4;
  byte flashLed1=false, flashLed4=false, flashTime;
  unsigned int mscnt=0;

    for (i=0; i<nleds; i++) {  // clear all the LED lites
      Lite = lites[i];  // get the HW adr of the LED
      digitalWrite(Lite,!ON_STATE);  // set it to the off state
    }

    while (!(btn1 && btn4)) {
      delay(5);
      mscnt+=5;

      lastB1 = btn1;
      lastB4 = btn4;
      //  determine curret buttons state
      btn1 = ! digitalRead(button[0]);  // button state is normally pulled high (so low = pressed)
      btn4 = ! digitalRead(button[3]); 

      if (btn1 && !lastB1) {
        flashLed1 = (flashLed1)? false : true;  // toggle the flashing
        flashLed4 = false;
      }
      if (btn4 && !lastB4) {
        flashLed4 = (flashLed4)? false : true;
        flashLed1 = false;
      }
      
      flashTime = (mscnt/400)%2; // changes state every ~400ms
      Lite = lites[0];  // control this LED
      if (btn1 || (flashLed1 && flashTime)) {
        digitalWrite(Lite, (ON_STATE));
      } else digitalWrite(Lite, (!ON_STATE));

      Lite = lites[nleds-1];  // control this LED
      if (btn4 || (flashLed4 && flashTime)) {
        digitalWrite(Lite, (ON_STATE));
      } else digitalWrite(Lite, (!ON_STATE));
    } 
}

void SuperComp60() {  // simulate a 1960's "Super Computer's display" by emulating HW flashing LEDs with slightly differing rates ~1Hz
  int flashRate[nleds];
  unsigned int myMS=0, shuffleID;
  int period = 1000;  // 1000 ms = 1Hz (~1/2 sec on & 1/2 sec off) has ~30 min. overall cycle
  
  shuffleID = micros(); // randomizing reference for the PRIG()
  clearDisp();
  // setup the mixed bag of LED flash rates
  for (int i=0; i<nleds; i++) {
    flashRate[i] = period + PRIG(i, shuffleID, 2*nleds);  // set pseudo random values, yet different unique values
  }

  while (!(ESC || btn4)) {    // press B4 ... B1 to restart
    if (btn1) myMS += 5;      // hurry up the process
    if (btn2Pressed) {               // goto all new random state
      myMS = random(999999);
      for (int i=0; i<nleds; i++) {
        flashRate[i] = period + PRIG(i, myMS, 2*nleds);
      }
    }

    for (int i=0; i<nleds; i++) {  // flash all the LEDs at their own rate
      lit[i+1] = (myMS%flashRate[i]) < (period/2);
    }
    myMS++;
    refreshWait(1);
    scanBtns();
  }
}

//================== major project funtional building blocks ===================


void PushIt() { // PushIt loop
  // Nov 2017 reworked 'PushIt' operation to be useful with displays of < 12 Leds
  bool lastPressed, pressed = false, pushed = false;

  static int maxCnt = 2; // starts at 2 after power on
  int litPeriod, period, cnt, lastMaxCnt, savedCnt;
  bool isLit, incDone = false, decDue = false;

  clearDisp();
  refreshDisp();

  // Push-It init code
  // --------------------------
  litPeriod = 1000;
  period = litPeriod;
  cnt = 0; // =1 resulted in initially one too few flashes

  // read previous incarnation of 'max count' from EEPROM for the 'Push-It' game
  savedCnt = EEPROM.read(eeCntAddr);
  maxCnt = (savedCnt < 50)? savedCnt: 2; // be reasonable

  showLevel(maxCnt);

  digitalWrite(ledOnBrd, HIGH);  //needs testing with RP2040
  t0 = millis();
  isLit = false;

  while (1) {
    pressed = btnPressed() && !btn4;
    if (!lastPressed && pressed)  {
      pushed = true;
    }

    // Flash from 1 to n times
    // Increasing the number every time the user has pressed the button on the first flash of the series
    // A missed attempt will reduce the count

    if ((millis() - t0) >= period) {  //    is this time period up ?
      if (cnt <= 1) {
        litPeriod = (nleds < 12) ? 800 : 1000;
        for (int i = 1; i < maxCnt; i++) {
          //good for 12 leds:    litPeriod = 16 * (litPeriod-100) / 20 +100;  //  P = 0.8*(P-100) + 100
          litPeriod = 15 * (litPeriod - 80) / 20 + 80; //  P = 0.75*(P-80) + 80
        }
      }

      if (isLit) {                       // about to go from LED on to OFF
        if (cnt == maxCnt + 1) { // on orig 'Push-It' this was if (cnt == 1) {
          //             !!!!!!!!!!!!!!!! ------------------------------------  time to judge
          lastMaxCnt = maxCnt;
          if (pushed) {       // he hit the mark
            maxCnt++;
            cell[lastMaxCnt] = 1;
            showLevel(maxCnt);  // acknowledge with 'showLevel' so we can discern levels above nleds
            refreshWait(500);   
            incDone = true;
            if (maxCnt > nleds) {
              refreshWait(500);
              taDah();       // reached the highest level
              //              maxCnt=2;   // now, allow levels > nleds
            }
          }

          if (maxCnt != lastMaxCnt) { Serial.print("maxCnt = "); Serial.println(maxCnt); }
          if (savedCnt != maxCnt && (maxCnt == lastMaxCnt)) {  // dont write more often than needed
            savedCnt = maxCnt;
            EEPROM.write(eeCntAddr, maxCnt);
          }

          cnt = maxCnt + 1;
        } else {
          // isLit and not time to judge
        }
        period = litPeriod / 3;

      } else {                     // about to go from LED off to ON
        period = litPeriod;
        cnt++;
        if (cnt > (maxCnt + 1)) { // at maxCnt LED is off for one count
          cnt = 1;             // ===========  restart the cycle
          incDone = false;
          digitalWrite(ledOnBrd, HIGH); // needs testing with RP2040
        }
      }
      t0 = millis();
      isLit = not isLit;

      if (cnt > 0 && cnt <= maxCnt) {  // && cnt < maxCnt
        if (pushed && !incDone) {
          maxCnt--;
          if (maxCnt < 1) maxCnt = 1;
          cell[maxCnt] = 0;
          brightOne = 0;
          beep(100);    // had a miss & a strike
        }
      }
      lastPressed = pressed;
      pushed = false;
    }

    //-------------- UPDATE Display
    litLED = (maxCnt - 1) % nleds + 1;
    if (litLED == 1) litLED = 0; // so you can tell when it flashes
    for (int i = 1; i <= nleds; i++) {
        dim[i] = (i <= maxCnt);
    }
    if (maxCnt > nleds) { // make it easier to see the fast flashes
      brightOne = 0;
      spriteTwo = 0;
      dim[1] = 0;
    }

    if (isLit && cnt < maxCnt + 1) spriteTwo = 1; // brighten led #1
    else spriteTwo = nleds + 1; //ledON(13); // No 'ON' lit but used up one's normal time, so dim leds don't flicker  ###TODO
    if (cnt == maxCnt + 1 && pushed) {
      spriteTwo = (maxCnt + 1);
    }

    refreshWait(1);
    if (ESC || btn4) {
      refreshWait(300);
#ifdef RP2040
      EEPROM.commit();  // use with RP2040
#endif
      return;
    }

    if (btn2) { // re initial game level
      maxCnt = 2;
      brightOne = 0;
      cnt = 0;
    }
  }
}

void Skee_ball() { // Skee ball
  // three target "hoops", worth (1) (3) (2)  pts. The closest is the 1pt hoop
  // You are scored after each 5 balls (15 is the maximum score)
  // hold Btn1 to swing, and let go to throw the ball
  int player = 0;
  int i, cnt;
  byte spot, cell, balls, score;
  byte h1, h2, h3;  // hoops valued 1-3
  int travel, rate, thrust;
  bool rolling, newPressed, released, btnDown, wasDown;

  loopCnt = 0;
  level = 1;
  ver = 1;  //if (ver==0) return;

  clearDisp(); refreshWait(500);
  while (1) {
    // setup game-level
    wait_BtnRelease();   // wait for no button pressed
    spot = 1;
    cell = 0;
    balls = 0;
    score = 0;

    clearDisp();
    lit[nleds - 2] = true;
    dimFlash = nleds - 2;
    dim[nleds - 3] = dim[nleds - 1] = true;

    h1 = nleds - 3; h3 = nleds - 2; h2 = nleds - 1;
    playing = true;
    while (playing) {
      spot = 1;       // where we are throwing from
      dimLED = 1;
      travel = 0;
      thrust = 0;
      rolling = false;
      turn = 1;   // # of rolled balls
      while (turn) {
        bool pressed = btnPressed();
        wasDown = btnDown; // keep B2&3 from wasing a 'release'
        btnDown = (btn1 || btn4);
        if (btnDown && !rolling) {
          thrust++;
          cell = 1;
          spriteOne = 1;
        }

        released = wasDown && (!btnDown) && BtnChanged;
        if (released) {   // the ball takes flight
          rolling = true;
          dimLED = 0;
          thrust = (3 * thrust) / (level + 2); // at higher levels targets are effectivly further, require more thrust
          //Serial.print("thrust: "); Serial.println(thrust);
          rate = (thrust + 750) / 15; //for thrust=7500 gives 100 which results in 10 spots in one second
          rate = rate / 4; // patch so we can update more often for smoother fast motion balls
        }

        if (rolling && (loopCnt % 50) == 0) { // was %200 before patch above
          if (thrust <= 30) {         // coming to a stop
            if ((nleds - 3) <= cell && cell <= (nleds - 1)) { // goes in one of the hoops
              score += (cell == h3) ? 3 : (cell == h2) ? 2 : 1;
              beep(40, 15);   // Tah
              if (cell == h2) beep(30);
              if (cell == h3)  {
                refreshWait(130);
                beep(200);   // Dah
              }
            } else {      // missed
              buZZ(50);
            }
            turn = 0;
            balls++;
            refreshWait(1000);
            cell = 0;
            spriteOne = 0;

            if (balls >= 5) {  // 5 ball set complete, give score
              showLevel(score);   // show the Score
              refreshWait(2000);
              if (score > 10) level++;
              playing = false;
              break;  // from this game-set
            }
            continue;
          }

          travel = travel + rate;
          thrust -= rate;
          cell = spot + travel / 100;
          if (cell >= nleds + 1) cell = (cell % nleds) + 1;
          spriteOne = cell;
        }

        //================ UPDATE Display
        loopCnt++;
        refreshWait(1);
        if (btn1 && btn4) return; // Only exit foe B1+4, so Btn4 can also be used to play
      }
    }
  }
}

// =====================================================================
void pingPong() { // Ping-Pong - Two player Game
 #define STARTPERIOD 330;
  int litPeriod, period;
  int cnt = 0, maxCnt;
  int rally = 0;
  int dir = 1;
  bool slam;
  bool isLit = false;
  bool lastSwung, swing = false;
  bool lastPosHit, hit = false;
  int ver;

  ver = 2;    // version 2: for 2 players P1 must use Btn1, P2: btn4
  //  ver = getOption(2); // version 1: use either Btn1 or 4 good for one player
  litPeriod = STARTPERIOD;
  period = litPeriod;
  maxCnt = nleds - 1;
  t0 = millis();

  while (1) {
    lastSwung = swing;
    scanBtns();
    if (ESC) return;
    swing =  btn1Pressed || btn4Pressed;
    if (ver == 2) {
      if (dir == 1) swing = btn4;
      else        swing = btn1;
    }
    hit = hit || swing;
    if (btn3Pressed) {    // step up the speed/level
      rally = rally + 1;
      litPeriod = STARTPERIOD;
      for (int i = 0; i < rally; i++) litPeriod = 9 * litPeriod / 10;
      cnt = (dir==1) ? 0 : nleds - 1; // restart the rally
      isLit = false;
    }

    if ((millis() - t0) >= period) {  // is this time period up ?
      if (isLit) {                       // about to go from LED on to OFF
        //if (cnt==5) {Serial.print("Period was: ");  Serial.println(period);} 
        period = litPeriod / 3;
          if (cnt >= maxCnt || cnt <= 0) { // ------------ end of the line
            litPeriod = STARTPERIOD;
            for (int i = 0; i < rally; i++) {
              litPeriod = 9 * litPeriod / 10;   // * .9
            }

            //        !!!!!!!!!!!!!!!! ------  make judgement
            if (hit ^ lastPosHit) {
              // -------------------  they hit it
              litLED = cnt + 1;
              //refreshWait(40);
              slam = lastPosHit;
              if (slam) litPeriod = litPeriod / 2; // a slam-hit
              boop(50 + slam*100);  // paddle hitting ball
              litLED = 0;
              hit = false;
              rally++;
              dir = -dir;
              cnt = (dir==1)? 0 : nleds - 1;
            } else {  // ------------  failed to hit it
              litLED = 0;
              clearDisp();
              // -----  acknowledge the "Win-ner!"
              if (rally>0) {
                if (dir > 0) {
                  brt[1] = 1;
                  beep(150,100);
                  brt[2] = 1;
                  beep(500);
                }
                else          {
                  brt[nleds - 1] = 1;
                  beep(150,100);
                  brt[nleds] = 1;
                  beep(500);
                }
              }

              refreshWait(1000);      // that round is over ...
              if (rally >= 2) {
                showLevel(rally / 2); // give em one star for each set of back & forth
                refreshWait(2000);
                brightOne = 0;
                refreshWait(200);
              }

              // re-init for a new round
              litPeriod = STARTPERIOD;
              period = litPeriod;
              rally = 0;
              cnt = (dir==1)? 0 : nleds - 1;
              clearDisp();
            }
          } // end time to judge
      } else {                     // about to go from LED off to ON
        period = litPeriod;
        lastPosHit = hit;
        hit = false;
        cnt = cnt + dir;  // goto NEXT cell
      }
      t0 = millis();
      isLit = not isLit;
    }

    //-----------  UPDATE Display
    for (int i = 1; i <= nleds; i++)  dim[i] = 0;
    dim[cnt + 1] = 1;
    refreshWait(1);
    if (ESC) return;
  }
}

//  a Simple implementation of a Simon Say game
void  SimpleSimon() {
  int i, n=1; // number of note in the sequence
  byte seq[50+1];

  // --------- Intro didy
  setColor(1); genTone(250,40); delay(4*40); 
  setColor(2); genTone(300,2*40); delay(40); setColor(3); genTone(300,40); delay(3*40);
  setColor(4); genTone(400,4*40);

  while (!ESC) {
    clearDisp(); refreshWait(500);
    // --------- select next Note Sequence challege, by adding a random note
    if (n==1) seq[n] = random(4)+1;
    seq[++n] = random(4)+1;

    // --------- Play the Sequence of Notes
    for (i=1; i<=n; i++) {
      delay(300);
      setColor(seq[i]);
      genTone(250+50*seq[i], 400);
    }
    clearDisp();

    // --------- listen and judge player's attempt to reproduce it
    i=1;
    while(!ESC) {
      refreshWait(1);
      if (btnPressed()) {    // TBD there is ambiguity here
        if (seq[i]==btnNum) {           // its good
          setColor(btnNum);
          genTone(250+50*btnNum, 400);
          if (++i > n) { // has completed the sequence
            if (n>=4) winner(1);
            break;
          }
        } else {                        // its a mistake
          noGood();
          if (n>4) showLevel(n-1);
          n=1;  // set for a new round
          break;
        }
      }
    }
    scanBtns(); // update btn status to be sure to catch B1+B4 for exit
  }
}

//----------------  React_Now - Swat Quickly as LEDs light up
void React_Now_ver2() { // attemp to hit the button when the first LED lights up
                   // 1st LED is lit for 100 msecs, additional LEDs lite every 20ms
                   // challenge: if one pushes & holds before the first LED they get a wrong 'win'; how to fix ?
//  bool btn = !PRESSED_STATE;
  int min_Period = 100;
//  int btnInput;
  int LEDptr;
  
//  btnInput = button[0];   // use the first button on the left of the four
  while(true) {
//    scanBtns();
    // insure that the button is not already pressed
    while (Btn && !ESC) scanBtns();
    if (ESC) return;
    
    // insure all LEDs are off
    clearDisp();

    delay(random(2000,5000));       // wait for 2 - 5 seconds
    lit[1]=1;
    refreshDisp();
    beep(10); // takes 2x msecs
    delay(min_Period-20-20);    // account for the 20ms beep time and 20ms in for loop

    for (LEDptr=1; LEDptr<=8; LEDptr++) {
//      digitalWrite(lites[litLED], ON_STATE);
//      delay(20);
      lit[LEDptr]=true;
      refreshWait(20);
      //scanBtns();
      stableBtns();  // scan for new button state
 
      if (BtnPressed) { // they pressed the button
        brightOne=LEDptr;
        refreshWait(3000);
        break;
      }
    }
    if (LEDptr>8) beep(100);
//    scanBtns();
    if (ESC) break;
  }
}

// ================ Generally Useful Support Utility Functions =============== 

// --------------------------------------------------------
// ------  Psydo Random Index Generator for selecting unique items from a set
// using a Miller Shuffle Algorithm (x_lite) see [github.com/RondeSC/Miller_Shuffle_Algo]
// produces a shuffled Index given a base Index 'inx', a random seed 'mixID' and the number of items being indexed
//
unsigned int PRIG(unsigned int inx, unsigned long mixID, unsigned short nlim) {
	const unsigned short p1 = 6719, p2 = 9973;  // maximum usable 'nlim' = 2*p2-1
	unsigned short si, r1, r2, rx;

	mixID += 131 * (inx / nlim);  // have inx overflow effect the mix
	si = (inx + mixID) % nlim;    // cut the deck

	r1 = mixID % p2;	// set of fixed randomizing values
	r2 = mixID % p1;
	rx = (mixID / nlim) % nlim + 1;

	// perform conditional multi-faceted mathematical spin-mixing  (average # of si recalcs performed is ~1.5)
	if (si % 3 == 0) si = (((si / 3) * p1 + r1) % ((nlim + 2) / 3)) * 3; // spin multiples of 3 
	if (si % 2 == 0) si = (((si / 2) * p2 + r2) % ((nlim + 1) / 2)) * 2; // spin multiples of 2 
	if (si < rx) si = (si * p1 + r2) % rx;                       // mix random half one way
	else         si = ((si - rx) * p2 + r1) % (nlim - rx) + rx;  // and the other another

	return(si);  // return 'Shuffled' index
}

// simular to 'showColor'. but does not use or effect std refreshDisp()
void setColor(byte color) {   // replaced 'allColor=' use
  bool on[9];                //  0: none, 1:RED, 2:YEL, 3:GRN, 4:BLUE
  onSet = false;
  if (color == 0) {
    refreshDisp();
    return;
  }
  for (int i = 1; i <= nleds; i++) on[i] = 0; // init working array
  if (color == 1 || color == 5)  on[1] = on[2] = true; // set Red LEDs
  if (color == 2 || color == 5)  on[3] = on[4] = true; // set Yel LEDs
  if (color == 3 || color == 5)  on[5] = on[6] = true; // set Blu LEDs
  if (color == 4 || color == 5)  on[7] = on[8] = true; // set Grn LEDs
  // -------------- turn the resulting Leds ON
  for (int i = 0; i < nleds; i++) {
    digitalWrite(lites[i], (on[i + 1] == ON_STATE));
  }
  onSet = true;
}



// -----------------------
void showLevel(byte level) {
  int n;
  clearDisp();
  refreshWait(250);
  for (int i=0; i<level; i++) {
    beep(10);
    n = (i%nleds)+1;  // n wraps around as i exceeds nleds
    dim[n]=true;
    brightOne=n;
    refreshWait(100);
  }
  refreshWait(750);
  clearDisp();
  refreshWait(250);  
}

// -----------------------   
byte get_Mode(byte modeCnt) {  // modeCnt is the max selectable mode #
  byte mode;
  byte ledSet=0;
  byte ledMask=0;
  int cylceTime=650;
  
  clearDisp();
  if (modeCnt==0 || modeCnt>4) {buZZ(500); return(-1);}
  for (int i=1; i<=(2*modeCnt); i++) {dim[i]=true;} // set relavent leds dim
  while(btnDown()) ;  // wait for button release
  while(! Btn || btnNum>modeCnt) {
    if (ESC) return(0);
    if (0==(msCnt%cylceTime)) {  // every 1/2 sec highlight a posible selection
      ledSet = (ledSet+1) % modeCnt;  // cycles 0 1 2 3 
      ledMask = 0x03 << 2*(ledSet);
      set_LEDs(ledMask,2);
    } if (200==(msCnt%cylceTime)) {
      set_LEDs(ledMask,1);  // undo last lit
    }
    refreshWait(1);
    scanBtns();
  }
  set_LEDs(ledMask,1);  // undo last lit
  mode = btnNum;
  // highlight bright their selection
  ledMask = 0x03 << 2*(mode-1);
  set_LEDs(ledMask,3);
  while(Btn) {
    refreshWait(1);
    scanBtns();
    if (ESC) return(0);
  }
  clearDisp();
  return mode;
}

// -----------------------   
byte get_Option(byte optCnt, int dflt) {  // optCnt is the max # option, dflt is the default value
  clearDisp();
  for (int i=1; i<=optCnt; i++) {dim[i]=true;}
  flashLed=dflt;
  while(btnDown()) ;  // wait for button release
  while(! btn1) {
      refreshWait(1);
      if (btnPressed()) {
        if (btn2Pressed) flashLed = (flashLed==0)?  optCnt : flashLed-1;  // -- move left Allow 0:debugging
        if (btn3Pressed) flashLed = (flashLed%optCnt) + 1;              // ++ moves right
        //if (btn4Pressed) flashLed = (flashLed%optCnt) + 1; // allow most play with only 2 buttons (1&4)
      }
      if (ESC) return(0);
  }
  //clearDisp(); this will reset flashLed !
  return(flashLed);
}

bool YesNo() {
  clearDisp();
  set_LEDs(0xC3,1);  // dimly lite the Grn & Red LEDs
  while(btnDown()) ;  // wait for button release
  while(! (btn1 || btn4)) {
      refreshWait(1);
      if((millis()/500)%2) { //  toggle every 500 ms
          set_LEDs(0xC0,2);
          set_LEDs(0x03,1);
      } else {
          set_LEDs(0x03,2);
          set_LEDs(0xC0,1);
      }
      scanBtns();
  }
  clearDisp();
  return(btn1);
}

// -----------------------   
bool btnDown() {
  refreshWait(1);
  scanBtns();
  return (Btn!=0);
}

void noBtnDelay(int msec) {
  while (!Btn && msec-->0) {
    //delay(1);  // preferred if ISR does LED refresh
    refreshWait(1);
    scanBtns();
  }
}

void wait_BtnRelease() {
  //while (btnState!=0 || btnChanged) {  // cares about keys except FuncKey
  while (Btn || BtnChanged) {
    delay(1);
    scanBtns();
    if (ESC) return;
  }
//  ESC=time2Escape=false;
}

// -----------------------   
void stableBtns() { // get a stable state of the buttons post debouncing
//  refreshWait(1);
  scanBtns();
  while(!btnStable) {
    refreshWait(1);
    scanBtns();
  }
}

// ------- Game sounds

void taDah() {
  setColor(0);
  setColor(5); beep(50);                    // Tah
  setColor(0); refreshDisp(); delay(150);  // ...
  setColor(5); beep(200);                   // Dah
  if (level > 0) { // for games using 'level' show it
    showLevel(level);
    while (!btnPressed()) {
      refreshDisp();
    }
  }
  setColor(0); refreshDisp();
}

void winner(int player) {    // acknowledge the "Win-ner!"
    clearDisp();
    refreshWait(100);
    if (player==1) {brt[2]=1; beep(150); delay(100); brt[1]=1; beep(500);}
    else          {brt[nleds-1]=1; beep(150); delay(100); brt[nleds]=1; beep(500);} 
    clearDisp();
}
void  lost()  {   // player has lost, give the 'Raspberry'
  showColor(0);  refreshWait(300);
  showColor(1);  buZZ(100);
  showColor(0);
  showColor(4);  LOUDbuZZ(300); // real RaZ
  showColor(0);  refreshWait(500);
}

void  noGood()  {   // 'Nah, na,na!' player messed up
  showColor(0);  buZZ(160);
  showColor(5);  refreshWait(80);
  showColor(0);  buZZ(60);
  showColor(5);  refreshWait(50);
  showColor(0);  buZZ(80);
  refreshWait(250);
}

void LOUDbeep(int delayms){  // works for both Buzzers & Speakers
  for (int i=0; i<delayms; i++) {         // wait for a delayms ms
    digitalWrite ( AUDIO_OUT, HIGH);
    digitalWrite ( AUDIO_NEG, LOW);
    delayMicroseconds(500);
    digitalWrite ( AUDIO_OUT, LOW);
    digitalWrite ( AUDIO_NEG, HIGH);
    delayMicroseconds(500);
  }
  //digitalWrite ( AUDIO_OUT, LOW);;
  digitalWrite ( AUDIO_NEG, LOW);
}  

void LOUDbuZZ(int delayms){  // works for both Buzzers & Speakers
  for (int i=0; i<delayms; i+=5) {       // wait for >= delayms ms
    LOUDbeep(2);
    if (random(2)) refreshDisp();  // flicker LEDs during LOUD buZZ
    digitalWrite ( AUDIO_OUT, HIGH);
    digitalWrite ( AUDIO_NEG, LOW);
    delayMicroseconds(1500);
    digitalWrite ( AUDIO_OUT, LOW);
    digitalWrite ( AUDIO_NEG, HIGH);
    delayMicroseconds(1500);
    if (random(2)) refreshDisp();  // flicker LEDs during LOUD buZZ
  }
  //digitalWrite ( AUDIO_OUT, LOW);;
  digitalWrite ( AUDIO_NEG, LOW);
}

// -------- support Musical notes
// ---------------------  tone() type Audio Output
//                        true tone audio with a Speaker;  fine with Passive buzzer;
//                        only almost, sort of, ok with Active Buzzer, due to Moiré interference patterns

//   Generate Tone frequencies without using "tone()" (saving 1500 program bytes & 21 ram bytes)
//     sounding almost as good
void genTone(int toneFreq, int delayms) {      // Generate a specified TONE
  byte pinState = 0;       // 0 turns it off
  byte mss = 0, cnt = 0;

  if (toneFreq == 0) {
    refreshWait(delayms);
    return;
  }
  pinMode(AUDIO_OUT, OUTPUT);

  int msCnt = ( (long) delayms * toneFreq) / 1000;
  int halfcycle = (1000000L / toneFreq) / 2;
  int nhCycles = 2 * msCnt; // keep calcs out of loop for accuracy sake
  if (nhCycles < 2) nhCycles = 2;
  int cms = 1000 / halfcycle; // calc cnt/ms for disp upd
  for (int i = 0; i < nhCycles; i++) {     // wait for a delayms ms
    if (volHigh) digitalWrite ( AUDIO_NEG, pinState); // set NEG lead to opposite of + lead
    pinState = 1 - pinState;
    digitalWrite ( AUDIO_OUT, pinState);
    delayMicroseconds(halfcycle);
    // may sound better & more accurate, if no LED refreshing is done ! or do each ~2ms, using min. CPU
    if (++cnt >= cms && (pinState == 0 || halfcycle>500)) {
      refreshDisp();
      cnt = 0;
    }
  }
  digitalWrite ( AUDIO_OUT, LOW);
  digitalWrite ( AUDIO_NEG, LOW);
}

// Game music stuff
const int basicKeyNotes[] =
{0, 131, 147, 165, 175, 196, 220, 247, 262, 294, 330, 349, 392, 440, 494, 523, 587, 659, 698, 784, 880, 988, 1047, 1175, 1319};
//    1   2   3   4   5   6   7   8   9   10  11  12  13  14  15  16  17  18  19  20  21  22  23  24
//                                do  re  mi  fa  so  la  ti  do
//    C3  D3  E3  F3  G3  A3  B3  C4  D4  E4  F4  G4  A4  B4  C5  D5  E5  F5  G5  A5  B5  C6  D6  E6

void playNote(int note, int delayms) {
  genTone(basicKeyNotes[note], delayms);
}

void playNote(int note) {
  playNote(note, 500);
}

// -------------- Diagnostics Tools & Testing Utilities ----------------
void btnColor() {  // ========================== Combined Button - LED exercise & Reporting
  byte coLvl[5] = {0, 0, 0, 0, 0}; // level for each color LED set
  byte col_leds[] = {0, 1, 2, 4, 8};
  byte col_pat[] = {0, 0b00000011, 0b00001100, 0b00110000, 0b11000000};
  byte colNum;

  clearDisp();

  while (1) {
    scanBtns();
    // ------------------- button reporting
    if (BtnChanged) {
      Serial.println(" ");
      if (btn1) Serial.print("[Btn1] "); else Serial.print("       ");
      if (btn2) Serial.print("[Btn2] "); else Serial.print("       ");
      if (btn3) Serial.print("[Btn3] "); else Serial.print("       ");
      if (btn4) Serial.print("[Btn4] "); else Serial.print("       ");
      //Serial.print(btnState);
      beep(15); // make a key press click, for up&down
    }
    if (ESC) Serial.print(" [ESC] ");

    if (BtnPressed) {  // code from Color_lights()
      colNum = btnNum;
      coLvl[colNum] = (coLvl[colNum] + 1) % 5;
      set_LEDs(col_pat[colNum], 0);  // clear old color set's levels, so not additive
      // update the lit state of each set of color LEDs
      switch (coLvl[colNum]) {
        case 1: set_LEDs(col_pat[colNum], 1); break;  // sets to dim level
        case 2:
          set_LEDs(col_pat[colNum], 2); // additively sets to mid lit level, requested color(s)
          break;
        case 3: set_LEDs(col_pat[colNum], 3); break;  // sets to HI lite level
        case 4:
          setColor(colNum); // overrides till cleared, does not relay on refreshDisp() resources
          setColor(0);  // clear setColor override, settings supported in refreshDisp will be seen again
          coLvl[colNum] = 0; // & reset lvl to 0
          break;
        default:  break;
      }
    }

    brightOne= (ESC)? nleds:0; 
    if (ESC && btn1) break;    // only leave Btn Reporting if btn1 is also held.
    refreshWait(2);  // extra time extends ESC timeout period (1.5 -> 3 secs)
  }
}
